# -*- coding: utf-8 -*-
"""Writing reads to files.
"""
import copy
import os
from subprocess import Popen, PIPE
from xphyle import xopen

class BatchWriter(object):
    """Wrapper for a string writer (e.g. FifoWriter) that improves performance
    by buffering a set number of reads and sending them as a single call to the
    string writer.
    
    Args:
        writer: The string writer to wrap. Must be callable with two arguments
            (read1 string, read2 string).
        batch_size: The size of the read buffer.
        lines_per_row: The number of lines used by each read for the specific
            file format (should be passed by the subclass in a
            super().__init__ call).
        linesep: The separator to use between each line (defaults to os.linesep)
    """
    def __init__(self, writer, batch_size, lines_per_row, linesep=os.linesep):
        self.writer = writer
        self.batch_size = batch_size
        self.lines_per_row = lines_per_row
        self.bufsize = batch_size * lines_per_row
        self.linesep = linesep
        self.read1_batch = self._create_batch_list()
        self.read2_batch = copy.copy(self.read1_batch)
        self.index = 0
    
    def _create_batch_list(self):
        """Create the list to use for buffering reads. Can be overridden, but
        must return a list that is of size ``batch_size * lines_per_row``.
        """
        return [None] * self.bufsize
    
    def __call__(self, read1, read2):
        """Add a read pair to the buffer. Writes the batch to the underlying
        writer if the buffer is full.
        
        Args:
            read1: read1 tuple (name, sequence, qualities)
            read2: read2 tuple
        """
        self.add_to_batch(*read1, self.read1_batch, self.index)
        self.add_to_batch(*read2, self.read2_batch, self.index)
        self.index += self.lines_per_row
        if self.index >= self.bufsize:
            self.flush()
    
    def add_to_batch(self, name, sequence, qualities, batch, index):
        """Add a read to the batch. Must be implemented by a subclass.

        Args:
            name, sequence, qualities: Read data.
            batch: The batch index.
            index: The item index.
        """
        raise NotImplementedError()
    
    def __enter__(self):
        return self
    
    def __exit__(self, exception_type, exception_value, traceback):
        if self.index > 0:
            self.flush()
        self.close()
    
    def flush(self):
        """Flush the current read buffers to the underlying string writer.
        
        Args:
            last: Is this the last call to flush? If not, a trailing linesep
                is written.
        """
        if self.index < self.bufsize:
            self.writer(
                self.linesep.join(self.read1_batch[0:self.index]),
                self.linesep.join(self.read2_batch[0:self.index]))
        else:
            self.writer(
                self.linesep.join(self.read1_batch),
                self.linesep.join(self.read2_batch))
        self.writer(self.linesep, self.linesep)
        self.index = 0
    
    def close(self):
        """Clear the buffers and close the underlying string writer.
        """
        self.read1_batch = None
        self.read2_batch = None
        self.writer.close()

class FastqWriter(BatchWriter):
    """BatchWriter implementation for FASTQ format.
    """
    def __init__(self, writer, batch_size):
        super(FastqWriter, self).__init__(writer, batch_size, 4)
    
    def _create_batch_list(self):
        return [None, None, '+', None] * self.batch_size
    
    def add_to_batch(self, name, sequence, qualities, batch, index):
        batch[index] = '@' + name
        batch[index+1] = sequence
        batch[index+3] = qualities

class StringWriter(object):
    """Interface for classes that write strings to files.
    """
    def __call__(self, read1_str, read2_str):
        """Write strings to a pair of files.
        
        Args:
            read1_str, read2_str: The strings to write.
        """
        raise NotImplementedError()
    
    def close(self):
        """Close the underlying files.
        """
        raise NotImplementedError()

class FifoWriter(StringWriter):
    """String writer that opens and writes to a pair of FIFOs in a non-blocking
    way. Each FIFO is written to by opening a subprocess in which stdin is
    piped through pv (with -B option to set max buffer size) to a FIFO.

    Args:
        file1: Path to the read1 FIFO
        file2: Path to the read2 FIFO
        buffer: Command line to program that accepts and buffers between stdin
            and stdout
        kwargs: Additional arguments to pass to Popen
    """
    def __init__(self, file1, file2, buffer='pv -B ', **kwargs):
        self.fifo1 = Popen(
            '{buffer} > {fifo}'.format(buffer=buffer, fifo=file1),
            stdin=PIPE, shell=True, universal_newlines=True, **kwargs)
        self.fifo2 = Popen(
            '{buffer} > {fifo}'.format(buffer=buffer, fifo=file2),
            stdin=PIPE, shell=True, universal_newlines=True, **kwargs)
    
    def __call__(self, read1_str, read2_str):
        self.fifo1.stdin.write(read1_str)
        self.fifo2.stdin.write(read2_str)
    
    def close(self):
        for fifo in (self.fifo1, self.fifo2):
            fifo.stdin.close()
            fifo.terminate()

class FileWriter(StringWriter):
    """String writer that opens and writes to a pair of files.
    
    Args:
        file1: Path to the read1 file
        file2: Path to the read2 file
        kwargs: Additional arguments to pass to the ``open`` call.
    """
    def __init__(self, file1, file2, **kwargs):
        self.file1 = xopen(file1, 'wt', **kwargs)
        self.file2 = xopen(file2, 'wt', **kwargs)
    
    def __call__(self, read1_str, read2_str):
        self.file1.write(read1_str)
        self.file2.write(read2_str)
    
    def close(self):
        for fileobj in (self.file1, self.file2):
            fileobj.close()



from setuptools import setup
import sys

requirements = []

if sys.version_info < (3, 3):
    sys.stdout.write("At least Python 3.3 is required.\n")
    sys.exit(1)

import versioneer

setup(
    name='srastream',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Utilities for streaming NGS reads from SRA accessions.',
    url='https://github.com/jdidion/srastream',
    author='John Didion',
    author_email='john.didion@nih.gov',
    license='Public Domain',
    packages = ['srastream'],
    scripts = ['bin/sra_dump'],
    install_requires = ['xphyle'],
    tests_require = ['pytest', 'pytest-cov'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: Public Domain',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
)



"""Create iterators over batches of reads from an SRA accession.
"""
from ngs import NGS
from ngs.Read import Read
# Import members of .utils and .writers to make them available from the
# top-level module.
# pylint: disable=wildcard-import
from .utils import *
from .writers import *
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

# TODO: [JD] Incoprorate Popen hacks from
# https://github.com/brentp/toolshed/blob/master/toolshed/files.py
# https://github.com/wal-e/wal-e/blob/master/wal_e/pipebuf.py

class SraReader(object):
    """Iterates through a read collection for a given accession number using
    the ngs-lib python bindings.
    
    Args:
        accn: The accession number
        batch_iterator: An iterator over indexes of batches to fetch. Typically,
            this is created using a :class:`srastream.utils.Batcher`.
        batcher_args: If `batch_iterator` is None, these arguments are used to
            create a Batcher.
    
    Examples:
        # Use as context manager
        with SraReader(accn, batch_size=1000) as reader:
            for reads in reader:
                print("\n".join(str(read) for read in reads))
        
        # Use manually
        batch_iterator = Batcher(batch_size=1000)
        reader = SraReader(accn, batch_iterator)
        reader.start()
        print("Reading {} reads from SRA accession {} ({})".format(
            reader.read_count, reader.accn, reader.run_name))
        try:
            for reads in reader:
                print("\n".join(str(read) for read in reads))
        finally:
            reader.close()
    """
    def __init__(self, accn, batch_iterator=None, **batcher_args):
        self.accn = accn
        self.batch_iterator = batch_iterator or Batcher(**batcher_args)
        self.read_collection = None
        self.run_name = None
        self.read_count = None
        self.frag_count = None
    
    def __enter__(self):
        self.start()
        return self
    
    def __exit__(self, exception_type, exception_value, traceback):
        self.finish()
    
    def __iter__(self):
        if self.read_collection is None:
            raise ValueError("Must call start() first")
        for _, start, size in self.batch_iterator(total=self.read_count):
            with self.read_collection.getReadRange(start + 1, size, Read.all) as read:
                for _ in range(size):
                    read.nextRead()
                    yield sra_reads(read)
    
    def start(self):
        """Open the read collection.
        """
        self.read_collection = NGS.openReadCollection(self.accn)
        self.run_name = self.read_collection.getName()
        self.read_count = self.read_collection.getReadCount()
        # grab the first read use it to determine whether the dataset
        # is single- or paired-end
        with self.read_collection.getReadRange(0, 1) as read:
            self.frag_count = len(sra_reads(read))
    
    def finish(self):
        """Close the read collection.
        """
        if self.read_collection is not None:
            self.read_collection.close()
            self.read_collection = None
    
    @property
    def name(self):
        return self.getName()
    
    @property
    def paired(self):
        if self.frag_count is None:
            raise ValueError("Must call start() first")
        return self.frag_count == 2
    
    def __getattr__(self, name):
        if self.read_collection is None:
            raise AttributeError(
                "'SraReader' has no attribute '{}'; you might need "
                "to call 'start()' first.".format(name))
        return getattr(self.read_collection, name)

def sra_reads(read, paired=None, expected_fragments=None):
    """Creates sequence of (name, sequence, qualities) tuples from the current
    read of an ngs.ReadIterator. Typically the sequence has one or two tuples
    for single- and paired-end reads, respectively.

    Args:
        read: an NGS.Read instance.
        paired: Whether this is paired-end data.
        expected_fragments: The number of fragments expected for this read. If
            None, fragment number is not validated.
    
    Returns:
        The tuple (frag1, frag2...), where each fragment is a tuple 
        (read_name, sequence, qualities).
    """
    read_name = read.getReadName()
    num_fragments = read.getNumFragments()
    if expected_fragments and num_fragments != expected_fragments:
        raise Exception("Read {} has fewer than {} fragments".format(
            read_name, expected_fragments))
    
    # TODO: extract other useful information such as read group
    #read_group = read.getReadGroup()
    
    def next_frag():
        """Create (name, bases, qualities) tuple from the next fragment.
        """
        read.nextFragment()
        if paired and not read.isPaired():
            raise Exception("Read {} is not paired".format(read_name))
        return (
            read_name,
            read.getFragmentBases(),
            read.getFragmentQualities())
    
    return tuple(next_frag() for i in range(num_fragments))

def sra_dump(
        accn, prefix=None, compression=True, fifos=False, batch_size=1000, 
        **batcher_args):
    """Convenience method to stream paired-end reads from SRA to FASTQ files.

    Args:
        accn: SRA accession.
        prefix: Output file prefix. If None, the accession is used.
        compression: Whether to compress the output files (bool), or the name of
             a compression scheme (e.g. 'gz', 'bz2', or 'xz').
        fifos: Whether output files should be FIFOs. If True, `compression` is
            ignored, and 'pv' must be callable. Can also be a string specifying 
            the program to use for buffering instead of pv.
        batcher_args: Specify arguments to the :class:`srastream.utils.Batcher`
            that will be used for batch iteration.
    
    Returns:
        A dict containing the output file names ('file1' and 'file2'),
        and read_count.
    """
    args = dict(
        ('file{}'.format(read), '{}.{}.fq'.format(prefix or accn, read)) 
        for read in (1,2))
    
    if fifos:
        if isinstance(fifos, str):
            args['buffer'] = fifos
        string_writer = FifoWriter(**args)
    else:
        if compression is True:
            compression = 'gz'
        if compression:
            args = dict(
                (key, '{}.{}'.format(name, compression)) 
                for key, name in args.items())
        string_writer = FileWriter(**args, compression=compression)
    
    reader = SraReader(accn, batch_size=batch_size, **batcher_args)
    writer = FastqWriter(string_writer, batch_size) 
    with reader, writer:
        for reads in reader:
            writer(*reads)
    
    args['accn'] = accn
    args['read_count'] = reader.read_count
    return args
